#!/usr/bin/perl

#        d s m M u l t i S e r v e r U p g r a d e . p l
#########################################################################
#
# This is a sample TSM for Space Management script for multi-server environments.
#
# The upgrade mechanism for managed file systems exists of three steps:
#
# 1. use the GPFS policy engine to generate a list of all migrated files
#    in the file system.
# 2. prepare the generated file list for the usage in the TSM command
#    dsmreconcile.
# 3. Invoke the dsmreconcile command with the arguments -upgrade -filelist
#    and perform the reconciliation.
#
# This script demonstrates how these steps can be implemented in perl to
# automate the whole reconiliation process.
#
#########################################################################

# set global variables ########################################################

$debug             = 2;

$nFiles            = 0;
$fsName            = 0;

$startTime     = time;
$tmpDir        = "/tmp";
$ruleFile      = $tmpDir."/dsmreconcile.gpfs.$startTime.policy.rule";
$ruleOutput    = $tmpDir."/dsmreconcile.gpfs.$startTime";
$prepOutput    = $tmpDir."/dsmreconcile.gpfs.$startTime.list.prep.out";
$reorgOutput   = $tmpDir."/dsmreconcile.gpfs.$startTime.list.reorg.out";
$scanOut       = "scan.out";

# get command line arguments ##################################################

$scriptName    = $0;

# declare sub routines ########################################################

sub verifyArguments
{
   die "Usage: $scriptName <filesystem>\n" if (@ARGV > 1 || @ARGV == 0);
   $fsName = $ARGV[0];
}

sub createRule
{
   open(RULE, ">$ruleFile")
   or die "Can't open $ruleFile: $!\n";
   
   print RULE "define(\n";
   print RULE "exlude_list,\n";
   print RULE "(PATH_NAME LIKE '%/.SpaceMan/%'\n";
   print RULE "OR NAME LIKE '%dsmerror.log%'\n";
   print RULE "OR NAME LIKE '.mmbackupShadow%')\n";
   print RULE ")\n";
   print RULE "RULE EXTERNAL LIST '$scanOut' EXEC ''\n";
   print RULE "RULE 'Reconcile' LIST '$scanOut'\n";
   print RULE "WHERE NOT (exlude_list)\n";
   print RULE "AND XATTR('dmapi.IBMServ') is NULL\n";
   close(RULE)
   or die "Can't close $ruleFile: $!\n";
}

sub invokePolicyEngine
{
   $policyCommand = "/usr/lpp/mmfs/bin/mmapplypolicy";
   $policyCommand = $policyCommand." $fsName -P $ruleFile -s $tmpDir";
   $policyCommand = $policyCommand." -f $ruleOutput -I defer";
   
   print "\n$policyCommand\n" if ($debug > 0);
   
   open(POLICY, "$policyCommand |")
   or die "Can't open $policyCommand: $!\n";
   
   while(<POLICY>)
   {
      print "$_" if ($debug > 1);
   }
   close(POLICY);
}

sub prepFilelist
{
   # given format from GPFS            : INODE IGEN SNAPID -- PATH
   # expected format from dsmreconcile : PATH

   # reorganize the line structure
   open(IN, "< $ruleOutput.list.$scanOut")
      or die "Can't open $ruleOutput.list.$scanOut: $!";
   open(OUT, ">$reorgOutput")
      or die "Can't open $reorgOutput: $!";
   while (<IN>)
   {
      print "$_\n" if ($debug > 2);
      if (/^(\d+) (\d+) \d+\s+ -- (.*)$/s)
      {
         print OUT "$3";
         $nFiles += 1;
      }
      else
      {
         print "Invalid filelist record: $_\n";
      }
   }
   close(IN);
   close(OUT);

   # sort the file list and change the \n to \0.
   $prepCommand = "sort $reorgOutput |";
   $prepCommand = $prepCommand." tr -s '\\n' '\\0' > $prepOutput";
   
   print "\n$prepCommand\n" if ($debug > 0);
   
   open(PREP, "$prepCommand |")
   or die "Can't open $prepCommand: $!\n";
   
   while(<PREP>)
   {
      print "$_" if ($debug > 1);
   }
   close(PREP);
}

sub invokeReconcile
{
   $dsmreconcileCommand = "dsmreconcile -upgrade -filelist=$prepOutput $fsName";
   
   print "\n$dsmreconcileCommand\n" if ($debug > 0);
   
   open(RECONCILE, "$dsmreconcileCommand |")
   or die "Can't open $dsmreconcileCommand: $!\n";
   
   print "\n";
   while(<RECONCILE>)
   {
      print "$_";
   }
   close(RECONCILE);
}

# run main programm ###########################################################

sub main
{
   &verifyArguments;
   
   print "\n## dsmreconcile starts at  : ".(localtime)." ####\n\n";
   print "## gpfs rule file          : $ruleFile\n";
   print "## scan output             : $ruleOutput.list.$scanOut\n";
   print "## reorganized output      : $reorgOutput\n";
   print "## prep output             : $prepOutput\n";
   
   print "\n## create policy rule      : ".(localtime)." ####\n";

   &createRule;

   print "\n## invoke policy engine    : ".(localtime)." ####\n";
   
   &invokePolicyEngine;
   
   print "\n## prepare / sort filelist : ".(localtime)." ####\n";
   
   &prepFilelist;
   
   print "\n## number of files         : ".$nFiles."\n";
   
   print "\n## invoke dsmreconcile     : ".(localtime)." ####\n";
   
   &invokeReconcile;
   
   print "\n## dsmreconcile ends at    : ".(localtime)." ####\n\n";
   
   if($debug == 0)
   {
      unlink $ruleFile;
      unlink $ruleOutput.list.$scanOut;
      unlink $reorgOutput;
      unlink $prepOutput;
   }
}

&main;





