#!/usr/bin/perl

#        d s m M u l t i S e r v e r R e m o v e . p l
#########################################################################
#
# This is a sample TSM for Space Management script for multi-server environments.
#
# The upgrade mechanism for managed file systems exists of multiple steps:
#
# 1. use the GPFS policy engine to generate a list of all migrated files
#    in the file system.
# 2. prepare the generated file list for the usage in the TSM command
#    dsmrecall.
# 3. Recall all files migrated to the specified TSM server.
# 4. prepare the generated file list for the usage in the TSM command
#    dsmreconcile.
# 5. Invoke the dsmreconcile command with the arguments -remove -filelist
#    and perform the reconciliation.
#
# This script demonstrates how these steps can be implemented in perl to
# automate the whole reconiliation process.
#
#########################################################################

# set global variables ########################################################

$debug             = 0;

$nFiles            = 0;
$fsName            = 0;
$serverName        = 0;
$useServerSpecRule = 1;

$startTime     = time;
$tmpDir        = "/tmp";

$reconcileRuleFile    = $tmpDir."/reconcile.gpfs.$startTime.policy.rule";
$reconcileRuleOutput  = $tmpDir."/reconcile.gpfs.$startTime";
$reconcilePrepOutput  = $tmpDir."/reconcile.gpfs.$startTime.list.prep.out";
$reconcileReorgOutput = $tmpDir."/reconcile.gpfs.$startTime.list.reorg.out";

$recallRuleFile       = $tmpDir."/recall.gpfs.$startTime.policy.rule";
$recallRuleOutput     = $tmpDir."/recall.gpfs.$startTime";
$recallPrepOutput     = $tmpDir."/recall.gpfs.$startTime.list.prep.out";

$cleanUpRuleFile      = $tmpDir."/cleanup.gpfs.$startTime.policy.rule";
$cleanUpRuleOutput    = $tmpDir."/cleanup.gpfs.$startTime";
$cleanUpPrepOutput    = $tmpDir."/cleanup.gpfs.$startTime.list.prep.out";
$cleanUpReorgOutput   = $tmpDir."/cleanup.gpfs.$startTime.list.reorg.out";
$scanOut              = "scan.out";

# get command line arguments ##################################################

$scriptName    = $0;

# declare sub routines ########################################################

sub verifyArguments
{
   die "Usage: $scriptName -server=<server> <filesystem>\n" if (@ARGV != 2);

   if($ARGV[0] =~ m/^-server=(.+)/)
   {
      $serverName = uc($1);
      die "ERROR : you need to specify a valid TSM Server name\n" if( length($serverName) == 0 );
   }
   else
   {
      die "Usage: $scriptName [-server=<server>] <filesystem>\n";
   }

   $fsName = $ARGV[1];
}

sub createReconcileRule
{
   open(RULE, ">$reconcileRuleFile")
   or die "Can't open $ruleFile: $!\n";
   
   print RULE "RULE EXTERNAL LIST '$scanOut' EXEC ''\n";
   print RULE "RULE 'Reconcile' LIST '$scanOut'\n";
   print RULE "    SHOW( XATTR('dmapi.IBMexID') )\n";
   print RULE "    WHERE XATTR('dmapi.IBMexID') IS NOT NULL\n";
   print RULE "    AND   XATTR('dmapi.IBMServ') LIKE 'server%'\n";
   
   close(RULE)
   or die "Can't close $ruleFile: $!\n";
}

sub createServerSpecRecallRule
{
   open(RULE, ">$recallRuleFile")
      or die "Can't open $recallRuleFile: $!\n";
   
   print RULE "define(is_migrated, (MISC_ATTRIBUTES LIKE '%V%'))\n\n";
   print RULE "define(server,$serverName)\n\n";
   print RULE "RULE EXTERNAL LIST '$scanOut' EXEC ''\n";
   print RULE "RULE 'Reconcile' LIST '$scanOut'\n";
   print RULE "    SHOW( XATTR('dmapi.IBMexID') )\n";
   print RULE "    WHERE XATTR('dmapi.IBMexID') IS NOT NULL\n";
   print RULE "    AND   XATTR('dmapi.IBMServ') LIKE 'server%'\n";
   print RULE "    AND   is_migrated\n";
   
   close(RULE)
      or die "Can't close $recallRuleFile: $!\n";
}

sub createCleanUpRule
{
   open(RULE, ">$cleanUpRuleFile")
      or die "Can't open $cleanUpRuleFile: $!\n";
   print RULE "define(is_premigrated, (MISC_ATTRIBUTES LIKE '%M%' AND MISC_ATTRIBUTES NOT LIKE '%V%'))\n\n";
   print RULE "define(server,$serverName)\n\n";
   print RULE "RULE EXTERNAL LIST '$scanOut' EXEC ''\n";
   print RULE "RULE 'Reconcile' LIST '$scanOut'\n";
   print RULE "    SHOW( XATTR('dmapi.IBMexID') )\n";
   print RULE "    WHERE XATTR('dmapi.IBMexID') IS NOT NULL\n";
   print RULE "    AND   XATTR('dmapi.IBMServ') LIKE 'server%'\n";
   print RULE "    AND   is_premigrated\n";
   
   close(RULE)
      or die "Can't close $cleanUpRuleFile: $!\n";
}

sub invokePolicyEngine
{
   my($ruleFile, $ruleOutput) = @_;
   $policyCommand = "/usr/lpp/mmfs/bin/mmapplypolicy";
   $policyCommand = $policyCommand." $fsName -P $ruleFile -s $tmpDir";
   $policyCommand = $policyCommand." -f $ruleOutput -I defer";
   
   print "\n$policyCommand\n" if ($debug > 0);
   
   open(POLICY, "$policyCommand |")
   or die "Can't open $policyCommand: $!\n";
   
   while(<POLICY>)
   {
      print "$_" if ($debug > 1);
   }
   close(POLICY);
}

sub prepRecallFilelist
{
   # given format from GPFS      : INODE IGEN SNAPID IBMEXOBJID -- PATH
   # expected format from recall : PATH

   $nFiles = 0;

   if( ! -f $recallRuleOutput.list.$scanOut ) 
   {
      open(OUT, ">>$recallRuleOutput.list.$scanOut")
      or die "Can't open $recallRuleOutput.list.$scanOut: $!";
      close(OUT);
   }

   # reorganize the line structure
   open(IN, "< $recallRuleOutput.list.$scanOut")
      or die "Can't open $$recallRuleOutput.list.$scanOut: $!";
   open(OUT, ">$recallPrepOutput")
      or die "Can't open $recallPrepOutput: $!";

   while (<IN>)
   {
      print "$_\n" if ($debug > 2);
      if (/^(\d+) (\d+) \d+\s+(\w+) -- (.*)$/s)
      {
         print OUT "$4";
         $nFiles += 1;
      }
      else
      {
         print "Invalid filelist record: $_\n";
      }
   }
   close(IN);
   close(OUT);
}

sub prepCleanUpFilelist
{
   # given format from GPFS            : INODE IGEN SNAPID IBMEXOBJID -- PATH
   # expected format from dsmreconcile : PATH

   $nFiles = 0;

   if( ! -f $cleanUpRuleOutput.list.$scanOut )
   {
      open(OUT, ">>$cleanUpRuleOutput.list.$scanOut")
      or die "Can't open $cleanUpRuleOutput.list.$scanOut: $!";
      close(OUT);
   }

   # reorganize the line structure
   open(IN, "< $cleanUpRuleOutput.list.$scanOut")
      or die "Can't open $cleanUpRuleOutput.list.$scanOut: $!";
   open(OUT, ">$cleanUpReorgOutput")
      or die "Can't open $cleanUpReorgOutput: $!";

   while (<IN>)
   {
      print "$_\n" if ($debug > 2);
      if (/^(\d+) (\d+) \d+\s+(\w+) -- (.*)$/s)
      {
         print OUT "$4";
         $nFiles += 1;
      }
      else
      {
         print "Invalid filelist record: $_\n";
      }
   }
   close(IN);
   close(OUT);
   
   # sort the file list and change the \n to \0.
   $prepCommand = "cat $cleanUpReorgOutput |";
   $prepCommand = $prepCommand." tr -s '\\n' '\\0' > $cleanUpPrepOutput";
   
   print "\n$prepCommand\n" if ($debug > 0);
   
   open(PREP, "$prepCommand |")
   or die "Can't open $prepCommand: $!\n";
   
   while(<PREP>)
   {
      print "$_" if ($debug > 1);
   }
   close(PREP);
}

sub prepReconcileFilelist
{
   # given format from GPFS            : INODE IGEN SNAPID IBMEXOBJID -- PATH
   # expected format from dsmreconcile : IBMEXOBJID INODE IGEN PATH

   $nFiles = 0;
   
   if( ! -f $reconcileRuleOutput.list.$scanOut ) 
   {
      open(OUT, ">>$reconcileRuleOutput.list.$scanOut")
      or die "Can't open $reconcileRuleOutput.list.$scanOut: $!";
      close(OUT);
   }
   
   # reorganize the line structure
   open(IN, "< $reconcileRuleOutput.list.$scanOut")
      or die "Can't open $reconcileRuleOutput.list.$scanOut: $!";
    open(OUT, ">$reconcileReorgOutput")
        or die "Can't open $reconcileReorgOutput: $!";
   while (<IN>) 
   {
      print "$_\n" if ($debug > 2);
      if (/^(\d+) (\d+) \d+\s+(\w+) -- (.*)$/s) 
      {
         print OUT "$3 $1 $2 $4";
         $nFiles += 1;
      } 
      else 
      {
         print "Invalid filelist record: $_\n";
      }
   }
   close(IN);
   close(OUT);
   
   # sort the file list and change the \n to \0.
   $prepCommand = "sort $reconcileReorgOutput |";
   $prepCommand = $prepCommand." tr -s '\\n' '\\0' > $reconcilePrepOutput";
   
   print "\n$prepCommand\n" if ($debug > 0);
   
   open(PREP, "$prepCommand |")
   or die "Can't open $prepCommand: $!\n";
   
   while(<PREP>)
   {
      print "$_" if ($debug > 1);
   }
   close(PREP);
}

sub invokeCleanUp
{
   $dsmreconcileCommand = "dsmreconcile -d -remove -filelist=$cleanUpPrepOutput $fsName";
   
   print "\n$dsmreconcileCommand\n" if ($debug > 0);
   
   open(RECONCILE, "$dsmreconcileCommand |")
   or die "Can't open $dsmreconcileCommand: $!\n";
   
   print "\n";
   while(<RECONCILE>)
   {
      print "$_";
   }
   close(RECONCILE);
}

sub invokeRecallServerSpec
{
   $dsmrecallCommand = "dsmrecall -servername=$serverName -filelist=$recallPrepOutput $fsName";
   
   print "\n$dsmrecallCommand\n" if ($debug > 0);
   
   open(RECALL, "$dsmrecallCommand |")
   or die "Can't open $dsmrecallCommand: $!\n";
   
   print "\n";
   while(<RECALL>)
   {
      print "$_";
   }
   close(RECALL);
}

sub invokeReconcile
{
   $dsmreconcileCommand = "dsmreconcile  -o -d -server=$serverName -filelist=$reconcilePrepOutput $fsName";
   
   print "\n$dsmreconcileCommand\n" if ($debug > 0);
   
   open(RECONCILE, "$dsmreconcileCommand |")
   or die "Can't open $dsmreconcileCommand: $!\n";
   
   print "\n";
   while(<RECONCILE>)
   {
      print "$_";
   }
   close(RECONCILE);
}

sub allowRemove
{
  $serverFile = "$fsName/.SpaceMan/multiserver/serverlist/$serverName";
  $removeFlag = "123\n";

  do {
    local $^I='.bak';		# in place editing
    local @ARGV=("$serverFile");
    while (<>) {
      # replace if present
      if ($. == 5)
      {
         print $removeFlag;
      }
      else
      {
         print $_;
      }
    }
    # append if missing
    if ($. == 4)
    {
      open(OUT, ">>$serverFile")
        or die "Can't open $serverFile: $!";
      print OUT $removeFlag;
      close(OUT);
    }
  }
}

# run main programm ###########################################################

sub main
{
   &verifyArguments;
   
   print "\n################################################################################\n";
   print "## recall of files starts at             : ".(localtime)." ####\n";
   print "################################################################################\n";
   print "## gpfs rule file                        : $recallRuleFile\n";
   print "## scan output                           : $recallRuleOutput.list.$scanOut\n";
   print "## prep output                           : $recallPrepOutput\n";
   print "\n## create policy rule for dsmrecall      : ".(localtime)." ####\n";
   &createServerSpecRecallRule;
   print "\n## invoke policy engine for dsmrecall    : ".(localtime)." ####\n";
   &invokePolicyEngine($recallRuleFile, $recallRuleOutput);
   print "\n## prepare / sort filelist for dsmrecall : ".(localtime)." ####\n";
   &prepRecallFilelist;
   print "\n## number of files                       : ".$nFiles."\n";
   print "\n## invoke dsmrecall                      : ".(localtime)." ####\n";
   &invokeRecallServerSpec;
   print "\n## dsmrecall ends at                     : ".(localtime)." ####\n\n";
   
   print "\n## sleep 60 seconds and let GPFS update the file attributes ####\n";
   sleep(60);

   print "\n################################################################################\n";
   print "## file attributes cleanup starts at     : ".(localtime)." ####\n";
   print "################################################################################\n";
   print "## gpfs rule file                        : $cleanUpRuleFile\n";
   print "## scan output                           : $cleanUpRuleOutput.list.$scanOut\n";
   print "## reorg output                          : $cleanUpReorgOutput\n";
   print "## prep output                           : $cleanUpPrepOutput\n";
   print "\n## create policy rule for cleanup        : ".(localtime)." ####\n";
   &createCleanUpRule;
   print "\n## invoke policy engine                  : ".(localtime)." ####\n";
   &invokePolicyEngine($cleanUpRuleFile, $cleanUpRuleOutput);
   print "\n## prepare / sort filelist               : ".(localtime)." ####\n";
   &prepCleanUpFilelist;
   print "\n## number of files                       : ".$nFiles."\n";
   print "\n## invoke dsmreconcile                   : ".(localtime)." ####\n";
   &invokeCleanUp;
   print "\n## file attributes cleanup ends at       : ".(localtime)." ####\n\n";
   
   print "\n## sleep 60 seconds and let GPFS update the file attributes ####\n";
   sleep(60);
   
   print "\n################################################################################\n";
   print "## reconcile starts at                   : ".(localtime)." ####\n";
   print "################################################################################\n";
   print "## gpfs rule file                        : $reconcileRuleFile\n";
   print "## scan output                           : $reconcileRuleOutput.list.$scanOut\n";
   print "## reorg output                          : $reconcileReorgOutput\n";
   print "## prep output                           : $reconcilePrepOutput\n";
   print "\n## create policy rule for reconcile      : ".(localtime)." ####\n";
   &createReconcileRule;
   print "\n## invoke policy engine                  : ".(localtime)." ####\n";
   &invokePolicyEngine($reconcileRuleFile, $reconcileRuleOutput);
   print "\n## prepare / sort filelist               : ".(localtime)." ####\n";
   &prepReconcileFilelist;
   print "\n## number of files                       : ".$nFiles."\n";
   print "\n## invoke dsmreconcile                   : ".(localtime)." ####\n";
   &invokeReconcile;
   print "\n## dsmreconcile ends at                  : ".(localtime)." ####\n\n";
   &allowRemove;

   if($debug == 0)
   {
      unlink $recallRuleFile if (-f $recallRuleFile);
      unlink $recallRuleOutput.list.$scanOut if (-f $recallRuleOutput.list.$scanOut);
      unlink $recallPrepOutput if (-f $recallPrepOutput);
      
      unlink $cleanUpRuleFile if (-f $cleanUpRuleFile);
      unlink $cleanUpRuleOutput.list.$scanOut if (-f $cleanUpRuleOutput.list.$scanOut);
      unlink $cleanUpReorgOutput if (-f $cleanUpReorgOutput);
      unlink $cleanUpPrepOutput if (-f $cleanUpPrepOutput);
      
      unlink $reconcileRuleFile if (-f $reconcileRuleFile);
      unlink $reconcileRuleOutput.list.$scanOut if (-f $reconcileRuleOutput.list.$scanOut);
      unlink $reconcilePrepOutput if (-f $reconcilePrepOutput);
      unlink $reconcileReorgOutput if (-f $reconcileReorgOutput);
   }
}

&main;





